package net.es.oscars.policy.config;

import net.es.oscars.policy.database.LTypeDAO;
import net.es.oscars.policy.database.OperatorDAO;
import net.es.oscars.policy.database.Rule;
import net.es.oscars.policy.database.RuleDAO;
import net.es.oscars.policy.database.RuleVar;

import org.apache.log4j.Logger;

import java.io.IOException;
import java.io.StreamTokenizer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

public class RuleSetParser implements RuleParser{
    private ConfigUtil configUtil;
    private Logger log;
    private RuleDAO ruleDAO;
    private LTypeDAO ltypeDAO;
    private OperatorDAO operatorDAO;
    private HashMap<String, RuleVar> varNameMap;
    
    static final public String[] tokenTypes = {"","XPATH_LIST", "XPATH_ENUM"};
    
    static final public List<String> stringOps = Arrays.asList(new String[]{
            "==", "!=","startsWith", "endsWith", "contains"});
    static final public List<String> numberOps = Arrays.asList(new String[]{
            "==", "!=","<",">","<=",">="});
    static final public List<String> booleanOps = Arrays.asList(new String[]{
            "==", "!="});
    static final public List<String> dateOps = Arrays.asList(new String[]{
            "==", "!=","<",">","<=",">="});
    
    public RuleSetParser(){
        this.log = Logger.getLogger(this.getClass());
    }
    
    public void init(ConfigUtil configUtil, RuleDAO ruleDAO,
        LTypeDAO ltypeDAO,  OperatorDAO operatorDAO){
        this.configUtil = configUtil;
        this.ruleDAO = ruleDAO;
        this.ltypeDAO = ltypeDAO;
        this.operatorDAO = operatorDAO;
    }
    
    public void parseRule(StreamTokenizer tokenizer, List<Rule> ruleStack, 
            String token, String normalizedToken, int tokenType, String ruleOp) throws PolicyParserException, IOException{
        
        String lType = null;
        String rType = null;
        String lValue = null;
        List<String> ops = null;
        if(this.configUtil.getVarParsers().containsKey(normalizedToken)){
            tokenizer.nextToken();
            token = ParserUtil.readToken(tokenizer);
            if(!":".equals(token)){
                throw new PolicyParserException("Missing token ':' at line " + tokenizer.lineno());
            }
            tokenizer.nextToken();
            token = ParserUtil.readToken(tokenizer);
            lValue = normalizedToken+":"+token;
            if(!this.varNameMap.containsKey(lValue)){
                throw new PolicyParserException("Undefined variable '" + 
                        lValue + "' at line " + tokenizer.lineno());
            }
            lType = "SPECIAL_VAR";
            ops = new ArrayList<String>();
            rType = this.configUtil.getVarParser(normalizedToken).getOpType(
                        this.varNameMap.get(lValue).getOutputType(), ops);
        }else if(this.configUtil.getStringFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? tokenTypes[tokenType-1] : "XPATH_NODE";
            rType = "STRING";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = stringOps;
        }else if(this.configUtil.getNumberFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? tokenTypes[tokenType-1] : "XPATH_NUMBER";
            rType = "NUMBER";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = numberOps;
        }else if(this.configUtil.getBooleanFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? tokenTypes[tokenType-1] : "XPATH_BOOLEAN";
            rType = "BOOLEAN";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = booleanOps;
        }else if(this.configUtil.getDateFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? tokenTypes[tokenType-1] : "XPATH_NUMBER";
            rType = "DATE";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = dateOps;
        }else if("action".equals(token)){
            lType = "ACTION";
            rType = "STRING";
            lValue = "*";
            ops = stringOps;
        }else{
            throw new PolicyParserException("Invalid token '" + token + "' at line " + tokenizer.lineno());
        }
        
        tokenizer.nextToken();
        String op = ParserUtil.readToken(tokenizer);
        if(op == null || (!ops.contains(op))){
            throw new PolicyParserException("Invalid operator '" + op + "' at line " + tokenizer.lineno());
        }
        tokenizer.nextToken();
        String value = ParserUtil.readToken(tokenizer);
        //adjust value
        if("ACTION".equals(lType) && this.configUtil.getActionMap().containsKey(value)){
            value = this.configUtil.getActionMap().get(value);
        }
        //validate date format
        if("DATE".equals(rType) && (!this.configUtil.validateDateField(value))){
            throw new PolicyParserException("Invalid date '" + value + "' at line " + tokenizer.lineno());
        }
        //check if special variable
        if(lValue.matches("#.+#")){
            lType = "SPECIAL_VAR";
            lValue = lValue.replaceAll("#","");
        }
        tokenizer.nextToken();
        if(!";".equals(ParserUtil.readToken(tokenizer))){
            throw new PolicyParserException("Missing ';' at line " + tokenizer.lineno());
        }
        
        Rule rule = new Rule();
        rule.setLType(ltypeDAO.queryByParam("name", lType));
        rule.setLValue(lValue);
        rule.setOperator(operatorDAO.findByTypeName(rType, op));
        rule.setRValue(value);
        Rule existingRule = ruleDAO.findMatchingRule(rule);
        if(existingRule != null){
            rule = existingRule;
        }else{
            ruleDAO.create(rule);
        }
        
        //look at last rule
        if(ruleStack.get(0) == null){
            ruleStack.remove(0);
            ruleStack.add(0, rule);
        }else {
            Rule combinedRule = ParserUtil.combineRule(ruleStack.get(0), rule, 
                    ruleOp, this.ltypeDAO, this.operatorDAO, this.ruleDAO);
            ruleStack.remove(0);
            ruleStack.add(0, combinedRule);
        }
        this.log.info("parseRule." + lType + ".end");
    }
    
    private String prepareXPath(String token, String normalizedToken, int tokenType) {
        String xPath = this.configUtil.getXPathMap().get(normalizedToken);
        if(tokenType == ParserUtil.POSITION_TOKEN){
            String[] parts = token.split("-");
            xPath = xPath.replace("#N#", parts[parts.length-1]);
        }
        
        return xPath;
    }

    public void destroy() {
        return;
    }

    /**
     * @return the varNameMap
     */
    public HashMap<String, RuleVar> getVarNameMap() {
        return this.varNameMap;
    }

    /**
     * @param varNameMap the varNameMap to set
     */
    public void setVarNameMap(HashMap<String, RuleVar> varNameMap) {
        this.varNameMap = varNameMap;
    }
}
